% Text Mining with MATLAB 2nd Edition
% www.http://www.textmininglab.net/

% CHAPTER 13: Content Analysis

disp('Example 13.1 -----------------------------------------------------');
clear                                                              %(13.1a)
load datasets_ch13a
whos                                                               %(13.1b)
pause;

disp('Example 13.2 -----------------------------------------------------');
summary(domain)                                                     %(13.2)
pause;
  
disp('Example 13.3 -----------------------------------------------------');
summary(polarity)                                                   %(13.3)
pause;

disp('Example 13.4 -----------------------------------------------------');
% computes the number of documents per category groups             %(13.4a)
type_TN = sum((domain=="transportation")&(polarity=="negative"));
type_TP = sum((domain=="transportation")&(polarity=="positive"));
type_FP = sum((domain=="financial")&(polarity=="positive"));
type_FN = sum((domain=="financial")&(polarity=="negative"));
doctypes = ["TRN-NEG","TRN-POS","FIN-POS","FIN-NEG"];
hf = figure(1); % generates the pie chart                          %(13.4b)
figname = 'Proportion of documents across domains and polarities';
set(hf,'Color',[1 1 1],'Name',figname);
pie3([type_TN,type_TP,type_FP,type_FN],ones(1,4),doctypes);
pause;

disp('Example 13.5 -----------------------------------------------------');
dims = 2; % defines the dimensionality of the reduced space         %(13.5)
cdist = pdist(tfidfmtx,'cosine'); % computes cosine distances
dmap = mdscale(cdist,dims,'Criterion','stress'); % computes the projections 
pause;

disp('Example 13.6 -----------------------------------------------------');
hf = figure(2); % creates the figure                               %(13.6a)
figname = 'Two-dimensional map of the dataset described in Figure 13.1'; 
set(hf,'Color',[1 1 1],'Name',figname); 
% gets indexes for each type of documents                          %(13.6b)
TN = (domain=="transportation")&(polarity=="negative");
TP = (domain=="transportation")&(polarity=="positive");
FN = (domain=="financial")&(polarity=="negative");
FP = (domain=="financial")&(polarity=="positive");
doctypes = ["TRN-NEG","TRN-POS","FIN-NEG","FIN-POS"];
% generates the plot                                               %(13.6c)
hp = plot(dmap(TN,1),dmap(TN,2)'.',dmap(TP,1),dmap(TP,2),'.',...
          dmap(FN,1),dmap(FN,2),'.',dmap(FP,1),dmap(FP,2),'.');
mc = 'rbrb'; ms = '*dvo'; % sets colors and shapes for the markers 
for k=1:length(hp)
    set(hp(k),'Marker',ms(k),'LineStyle','none','MarkerSize',5);
    set(hp(k),'MarkerFaceColor',mc(k),'MarkerEdgeColor',mc(k));
end
% adjusts axis and prints legend and labels                        %(13.6d)
axis([-0.5,0.5,-0.6,0.5]);
legend(doctypes,'Location','South','Orientation','horizontal');
xlabel('Domain discriminating dimension')
ylabel('Polarity discriminating dimension');
pause;

disp('Example 13.7 -----------------------------------------------------');
% preprocesses and prepares the 100 positive reviews                (13.7a)
dirname = "review_polarity/txt_sentoken/pos/";
filelist = dir(dirname); counter = 0;
for k=1:length(filelist)
    if filelist(k).isdir==0
        counter = counter+1;
        %filename = filelist(k).name;
        %temp = string(evalc('type(dirname+filename)'));
        temp = string(evalc('type(dirname+filelist(k).name)'));
        tempdocs(counter) = strtrim(regexprep(temp,'\s*',' '));
    end
end
docs_pos = erasePunctuation(tokenizedDocument(tempdocs));
% preprocesses and prepares the 100 negative reviews                (13.7b)
dirname = "review_polarity/txt_sentoken/neg/";
filelist = dir(dirname); counter = 0;
for k=1:length(filelist)
    if filelist(k).isdir==0
        counter = counter+1;
        %filename = filelist(k).name;
        %temp = string(evalc('type(dirname+filename)'));
        temp = string(evalc('type(dirname+filelist(k).name)'));
        tempdocs(counter) = strtrim(regexprep(temp,'\s*',' '));
    end
end
docs_neg = erasePunctuation(tokenizedDocument(tempdocs));
pause;

disp('Example 13.8 -----------------------------------------------------');
hf = figure(3); % creates the figure                               %(13.8a)
set(hf,'Color',[1 1 1],'Name','Distributions of document lenghts');
% computes and plots the histogram for positive reviews            %(13.8b)
%for k=1:length(data_pos), lpos(k) = length(data_pos(k).token); end;
lpos = doclength(docs_pos);
subplot(2,1,1); hist(lpos,min(lpos):max(lpos));  
axis([0,2500,0,8]); text(1800,7,'Positive Reviews');
xlabel('Document length (in words)'); ylabel('Frequency');
% computes and plots the histogram for negative reviews            %(13.8c)
%for k=1:length(data_neg), lneg(k) = length(data_neg(k).token); end;
lneg = doclength(docs_neg);
subplot(2,1,2); hist(lneg,min(lneg):max(lneg));  
axis([0,2500,0,8]); text(1800,7,'Negative Reviews');
xlabel('Document length (in words)'); ylabel('Frequency');
pause;

% GENERATES TABLE 13.1 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
qp = docs_pos; qn = docs_neg; qx = [docs_pos,docs_neg];
qlp = lpos; qln = lneg; qlx = doclength(qx);
Positive = [length(qp);sum(qlp);length(qp.Vocabulary);min(qlp);max(qlp);mean(qlp)];
Negative = [length(qn);sum(qln);length(qn.Vocabulary);min(qln);max(qln);mean(qln)];
Collection = [length(qx);sum(qlx);length(qx.Vocabulary);min(qlx);max(qlx);mean(qlx)];
tbl = table([Positive,Negative,Collection]);
statnames = ["Number of Docs";"Running Words";"Vocabulary Size";...
             "Min Doc Length";"Max Doc Length";"Aver Doc Length"];
addvars(tbl,statnames,'NewVariableNames','Stats','Before',1)
pause;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 13.9 -----------------------------------------------------');
rng('default');                                                     %(13.9)
randselect_pos = randperm(1000);
randselect_neg = randperm(1000);
pause;

disp('Example 13.10 ----------------------------------------------------');
% NOT A SCRIPT                                                     %(13.10)
pause;

disp('Example 13.11 ----------------------------------------------------');
% gets the test documents for both categories                      %(13.11)
tst_docs_pos = docs_pos(randselect_pos(1:100));
tst_docs_neg = docs_neg(randselect_neg(1:100));
pause;

disp('Example 13.12 ----------------------------------------------------');
load datasets_ch13b % loads the models                            %(13.12a)
% computes the log likelihood ratios                              %(13.12b)
loglirat_pos = compute_loglirat(tst_docs_pos,unigram_pos,unigram_neg);
loglirat_neg = compute_loglirat(tst_docs_neg,unigram_pos,unigram_neg);
pause;

% GENERATES FIGURE 13.4 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
hf = figure(4); set(hf,'Color',[1 1 1],'Name','Log likelihood ratios for test set documents in both categories');
subplot(2,1,1); plot([1:length(loglirat_pos)],loglirat_pos,'b.-',[1,length(loglirat_pos)],[0,0],'r--')
xlabel('test documents in the positive category'); ylabel('log likelihood ratios')
subplot(2,1,2); plot([1:length(loglirat_neg)],loglirat_neg,'b.-',[1,length(loglirat_neg)],[0,0],'r--')
xlabel('test documents in the negative category'); ylabel('log likelihood ratios')
pause
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 13.13 ----------------------------------------------------');
% sets both the estimated and the actual category vectors         %(13.13a)
targets = [ones(length(loglirat_pos),1);zeros(length(loglirat_neg),1)];
estimated = double([loglirat_pos>0;loglirat_neg>0]);
% computes the confusion matrix                                   %(13.13b)
confusion_mtx = confusionmat(estimated,targets,'ORDER',[1,0]);
cmtx = array2table(confusion_mtx,'VariableNames',["positive";"negative"]);
cmtx.('Classified as') = ["positive";"negative"];
disp(cmtx)
% computes the overall accuracy                                   %(13.13c)
accuracy = sum(diag(confusion_mtx))/sum(sum(confusion_mtx))*100
pause;

disp('Example 13.14 ----------------------------------------------------');
% inputs the sample sentences in the required format              %(13.14a)
newdata(1) = "the movie is terrible";
newdata(2) = "the actor performance is excellent";
newdata(3) = newdata(2)+" but "+newdata(1);
newdocs = tokenizedDocument(newdata);
% estimates the polarity intensity                                %(13.14b)
loglirat = compute_loglirat(newdocs,unigram_pos,unigram_neg)'
pause;

disp('Example 13.15 ----------------------------------------------------');
% NOT A SCRIPT
pause;

disp('Example 13.16 ----------------------------------------------------');
scores = 2*(1+exp(-loglirat)).^(-1)-1                              %(13.16)
pause;

% Generates Table 13.2 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
strtable2(1) = "It was as good as garbage.";
strtable2(2) = "This actor is terrible; his performance was pathetic.";
strtable2(3) = "The music was bad, and the script was boring.";
strtable2(4) = "This film has some problems with the plot.";
strtable2(5) = "Not as good as the previous movie in the saga.";
strtable2(6) = "Not so bad. I was expecting something worse.";
strtable2(7) = "The plot was simple, but I enjoyed the movie anyway.";
strtable2(8) = "Interesting film, which is full of action and excitement.";
strtable2(9) = "A very funny and pleasantly entertaining film.";
strtable2(10) = "Wonderful script and beautiful photography. A great movie!";
strtable2(11) = "Excellent movie, as expected from such a great director.";
strtable2(12) = "Exceptional production I will be watching again and again.";
doctable2 = lower(erasePunctuation(tokenizedDocument(strtable2)));
logtable2 = compute_loglirat(doctable2,unigram_pos,unigram_neg);
scrtable2 = 2*(1+exp(-logtable2)).^(-1)-1;
poltable2 = strings(length(scrtable2),1);
poltable2(scrtable2>0)="Positive"; poltable2(scrtable2<0)="Negative";
table2 = table();
table2.loglirat = logtable2;
table2.polarity = poltable2;
table2.score = scrtable2;
table2.Example = strtable2';
disp(table2)
pause;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 13.17 ----------------------------------------------------');
% computes polarity intensity scores for a given list of words    %(13.17a)
words = ["awful","bad","excellent","good","hilarious","normal",...
         "perfect","poor","regular","terrible"];
loglirat_words = compute_loglirat(words,unigram_pos,unigram_neg);
scores_words = 2*(1+exp(-loglirat_words)).^(-1)-1;
% creates the bar graph                                           %(13.17b)
hf = figure(5); set(hf,'Color',[1 1 1]); ha = axes;
set(hf,'Name','Intensity score estimates at the word level');
barh(scores_words); xlabel('Intensity of Polarity');
axis([-1,1,0,11]); set(ha,'YTickLabel',words,'YDir','reverse'); 
pause;

disp('Example 13.18 ----------------------------------------------------');
% copies all negative documents into a single string               %(13.18)
strings_neg = joinWords(docs_neg);
collection_neg = join(strings_neg);
pause;

disp('Example 13.19 ----------------------------------------------------');
pattern1 = "this movie was (\w+) ";                                %(13.19)
pause;

disp('Example 13.20 ----------------------------------------------------');
matches = regexp(collection_neg,pattern1,'tokens');                %(13.20)
matches1_neg = string(matches)
pause;

disp('Example 13.21 ----------------------------------------------------');
% computes probability ratios for negative vocabulary terms       %(13.21a)
for k=1:length(unigram_neg.vocab)
    p_neg = unigram_neg.prob(k);
    idx = find(unigram_pos.vocab==unigram_neg.vocab(k));
    if isempty(idx), p_pos = unigram_pos.unkp;
    else, p_pos = unigram_pos.prob(idx); end
    score_neg(k) = p_neg/p_pos;
end
% ranks negative vocabulary according to probability ratios       %(13.21b)
[~,all_ranks_neg] = sort(score_neg,'descend');
pause;

disp('Example 13.22 ----------------------------------------------------');
ranks1_neg = zeros(1,length(matches1_neg));                        %(13.22)
for k=1:length(matches1_neg)
    idx = find(unigram_neg.vocab==matches1_neg(k));
    if isempty(idx), ranks1_neg(k) = mean(all_ranks_neg);
    else, ranks1_neg(k) = find(all_ranks_neg==idx); end
end
pause;

disp('Example 13.23 ----------------------------------------------------');
hf = figure(6); set(hf,'Color',[1 1 1]);                           %(13.23)
set(hf,'Name','Ranking of words matching the pattern "this movie was _"');
ha = axes; hp = plot(ranks1_neg,1:length(ranks1_neg),'o-');
set(hp,'MarkerFaceColor','b','MarkerSize',5);
axis([0,25000,0,length(ranks1_neg)+1]); xlabel('Rank index');
set(ha,'YTick',1:length(ranks1_neg),'YTickLabel',matches1_neg);
set(ha,'YDir','reverse','YGrid','on');
pause;

disp('Example 13.24 ----------------------------------------------------');
unique(matches1_neg(ranks1_neg<10000))                             %(13.24)
pause;

disp('Example 13.25 ----------------------------------------------------');
intersect(matches1_neg,unigram_neg.vocab(all_ranks_neg(1:10000)))  %(13.25)
pause;

disp('Example 13.26 ----------------------------------------------------');
% copies all positive documents into a single string              %(13.26a)
strings_pos = joinWords(docs_pos);
collection_pos = join(strings_pos);
% matches regular expression 'pattern1' from (13.19)              %(13.26b)
matches = regexp(collection_pos,pattern1,'tokens');
matches1_pos = string(matches)
pause;

disp('Example 13.27 ----------------------------------------------------');
% computes probability ratios for positive vocabulary terms       %(13.27a)
for k=1:length(unigram_pos.vocab)
    p_pos = unigram_pos.prob(k);
    idx = find(unigram_neg.vocab==unigram_pos.vocab(k));
    if isempty(idx), p_neg = unigram_neg.unkp;
    else, p_neg = unigram_neg.prob(idx); end
    score_pos(k) = p_pos/p_neg;
end
% ranks positive vocabulary according to probability ratios   	  %(13.27b)
[~,all_ranks_pos] = sort(score_pos,'descend');
% filters words matched in (13.26b) with the ranked vocabulary    %(13.27c)
intersect(matches1_pos,unigram_pos.vocab(all_ranks_pos(1:10000)))
pause;

disp('Example 13.28 ----------------------------------------------------');
word_sequence = "(?:this|that|the) (?:movie|film) (?:is|was)"; 	   %(13.28)
optional_word = "(?: \w*?y\>)?";
candidate_word = " (\w+) ";
pattern2 = word_sequence+optional_word+candidate_word; 
pause;

disp('Example 13.29 ----------------------------------------------------');
% extracts candidate words from the positive polarity subset      %(13.29a)
matches2_pos = string(regexp(collection_pos,pattern2,'tokens'));
% extracts candidate words from the negative polarity subset      %(13.29b)
matches2_neg = string(regexp(collection_neg,pattern2,'tokens'));
% assigns ranks to positive polarity candidate words              %(13.29c)
ranks2_pos = zeros(1,length(matches2_pos));
for k=1:length(matches2_pos)
    idx = find(unigram_pos.vocab==matches2_pos(k));
    if isempty(idx), ranks2_pos(k) = mean(all_ranks_pos);
    else, ranks2_pos(k) = find(all_ranks_pos==idx); end
end
% assigns ranks to negative polarity candidate words              %(13.29d)
ranks2_neg = zeros(1,length(matches2_neg));
for k=1:length(matches2_neg)
    idx = find(unigram_neg.vocab==matches2_neg(k));
    if isempty(idx), ranks2_pos(k) = mean(all_ranks_neg);
    else, ranks2_neg(k) = find(all_ranks_neg==idx); end
end
pause;

% Generates Figure 13.7 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
hf = figure(7); set(hf,'Color',[1 1 1]);         
set(hf,'Name','Ranking of words matching the pattern in (12.28)');
ha = subplot(1,2,1); plot(ranks2_pos,1:length(ranks2_pos),'.-');
axis([0,30000,0,length(ranks2_pos)+1]); xlabel('Rank index');
ylabel('Positive-polarity candidate words'); set(ha,'YDir','reverse');
ha = subplot(1,2,2); plot(ranks2_neg,1:length(ranks2_neg),'.-');
axis([0,30000,0,length(ranks2_neg)+1]); xlabel('Rank index');
ylabel('Negative-polarity candidate words'); set(ha,'YDir','reverse');
pause;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Example 13.30 ----------------------------------------------------');
% selects top-ranked positive candidate words                     %(13.30a)
selected2_pos = unique(matches2_pos(ranks2_pos<10000));
whos matches2_pos selected2_pos
% selects top-ranked negative candidate words                     %(13.30b)
selected2_neg = unique(matches2_neg(ranks2_neg<10000));
whos matches2_neg selected2_neg
pause;

disp('Example 13.31 ----------------------------------------------------');
disp(selected2_pos(1:10))                                         %(13.31a)
disp(selected2_neg(1:10))                                         %(13.31b)
pause;

disp('Example 13.32 ----------------------------------------------------');
details=tokenDetails(addPartOfSpeechDetails([docs_pos,docs_neg])); %(13.32)
adjectives=unique(details.Token(details.PartOfSpeech=='adjective'));
pause;

disp('Example 13.33 ----------------------------------------------------');
[selected2_pos,~,idx_pos]=unique(matches2_pos(ranks2_pos<10000)); %(13.33a)
counts_pos = hist(idx_pos,length(selected2_pos));
frequent_pos = selected2_pos(counts_pos>1);
[selected2_neg,~,idx_neg]=unique(matches2_neg(ranks2_neg<10000)); %(13.33b)
counts_neg = hist(idx_neg,length(selected2_neg));
frequent_neg = selected2_neg(counts_neg>1);
pause;

disp('Example 13.34 ----------------------------------------------------');
frequent_pos = setdiff(frequent_pos,stopWords);                   %(13.34a)
filtered2_pos = intersect(frequent_pos,adjectives)'
frequent_neg = setdiff(frequent_neg,stopWords);                   %(13.34b)
filtered2_neg = intersect(frequent_neg,adjectives)'    
pause;

disp('Example 13.35 ----------------------------------------------------');
% positive review words that are relevant to the negative category (13.35a)
disp(unique(matches2_pos(ranks2_pos>29000)))
% negative review words that are relevant to the positive category (13.35b)
disp(unique(matches2_neg(ranks2_neg>27000)))
pause;

disp('Example 13.36 ----------------------------------------------------');
candidate_word = " (\w+) ";                                        %(13.36)
qualifiers = unique([filtered2_pos,filtered2_neg]); 
anyqualifier = " (?:"+join(qualifiers,'|')+") ";
pattern3 = "(?:this|that|the)"+candidate_word+"(?:is|was)"+anyqualifier;
pause;

disp('Example 13.37 ----------------------------------------------------');
collection_all = join([collection_pos,collection_neg]);           %(13.37a)
matches3_asp = regexp(collection_all,pattern3,'tokens');
selected3_asp = unique(string(matches3_asp));
whos selected3_asp                                                %(13.37b)
pause;

disp('Example 13.38 ----------------------------------------------------');
nouns = unique(details.Token(details.PartOfSpeech=='noun'));      %(13.38a)
[selected3_asp,~,idx_asp] = unique(string(matches3_asp));         %(13.38b)
counts_asp = hist(idx_asp,length(selected3_asp));
frequent_asp = selected3_asp(counts_asp>1);
frequent_asp = setdiff(frequent_asp,stopWords);                   %(13.38c)
filtered3_asp = intersect(frequent_asp,nouns)'
pause;

disp('Example 13.39 ----------------------------------------------------');
clear; load datasets_ch13c                                        %(13.39a)
whos                                                              
keywds([5,25,35]) % some examples of the extracted keywords       %(13.39b)
pause;

disp('Example 13.40 ----------------------------------------------------');
details = tokenDetails(tokdoc);                                   %(13.40a)
size(details)
string(details.Properties.VariableNames)                          %(13.40b)
pause;

disp('Example 13.41 ----------------------------------------------------');
pattern1 = "(\w+ , )+(\w+ )(?:and |, and |or |, or )(\w+ )";       %(13.41)
pause

disp('Example 13.42 ----------------------------------------------------');
counter = 0;                                                      %(13.42a)
for k = 1:length(tokdoc)
    sentence = lower(joinWords(tokdoc(k)));
    matches = regexp(sentence,pattern1,'tokens');
    if isempty(matches), continue; end
    items = split(join(string(matches{1})),[" ",","]);
    items = items(strlength(items)>0);
    items = setdiff(items,stpwds);
    counter = counter+1;
    itemlists(counter).sentn = k;
    itemlists(counter).items = items';
end
struct2table(itemlists)                                           %(13.42b)
pause;

disp('Example 13.43 ----------------------------------------------------');
for k=1:length(itemlists)                                          %(13.43)
    sentn = itemlists(k).sentn;
    items = itemlists(k).items;
    fprintf('* In sentence #%d: %s\n',sentn,join(items,', '));
end
pause;

disp('Example 13.44 ----------------------------------------------------');
% collects common and proper nouns in the document                %(13.44a)
mask = (details.PartOfSpeech=='noun') |...
       (details.PartOfSpeech=='proper-noun');
nouns = unique(details.Token(mask));
nouns = nouns(strlength(nouns)>2); % keeps strings of length > 2
nouns = setdiff(lower(nouns),stpwds); % removes stop words
% creates a list of valid entities by combining nouns and keywds  %(13.44b)
entity = union(nouns,keywds);
pause;

disp('Example 13.45 ----------------------------------------------------');
entwrd = "("+join(entity,'|')+")";                                 %(13.45)
relwrd = "(?:like|are|is|such as|including)";
pattern2 = entwrd+"\W+?"+relwrd+"\W+?"+entwrd;
pause;

disp('Example 13.46 ----------------------------------------------------');
counter = 0;                                                      %(13.46a)
for k = 1:length(tokdoc)
    sentence = lower(joinWords(tokdoc(k)));
    matches = regexp(sentence,pattern2,'tokens');
    if isempty(matches), continue; end
    counter = counter+1;
    catinspairs(counter).sentn = k;
    catinspairs(counter).category = matches{1}(1);
    catinspairs(counter).instance = matches{1}(2);
end
struct2table(catinspairs)                                         %(13.46b)
pause;

disp('Example 13.47 ----------------------------------------------------');
for k=1:length(catinspairs)                                        %(13.47)
    cat = catinspairs(k).category;
    ins = catinspairs(k).instance;
    for n=1:length(itemlists)
        itm = itemlists(n).items;
        if sum(itm==ins)>0
            fprintf('* %s -> %s\n',cat,join(itm,', '));
        end
    end
end
pause;

disp('Example 13.48 ----------------------------------------------------');
relations = ["caused by","caused","causes","cause","contain",...   %(13.48)
             "contains","contained in","contained","provide",...
             "provides","provided by","provided","consists of",...
             "consists","involves","involve"];
relwrd = "("+join(relations,'|')+")";
pattern3 = entwrd+"\W*?(?:\w+ ){0,1}?"+relwrd+".*?"+entwrd;
pause;

disp('Example 13.49 ----------------------------------------------------');
counter = 0;                                                      %(13.49a)
for k = 1:length(tokdoc)
    sentence = lower(joinWords(tokdoc(k)));
    matches = regexp(sentence,pattern3,'tokens');
    if isempty(matches), continue, end
    counter = counter+1;
    entreltriples(counter).sentn = k;
    entreltriples(counter).entity1 = matches{1}(1);
    entreltriples(counter).relation = matches{1}(2);
    entreltriples(counter).entity2 = matches{1}(3);
end
struct2table(entreltriples)                                       %(13.49b)
pause;

disp('Example 13.50 ----------------------------------------------------');
disp(joinWords(tokdoc(43)))                                       %(13.50)
pause;

disp('Example 13.51 ----------------------------------------------------');
disp(joinWords(tokdoc(12)))                                       %(13.51a)
disp(joinWords(tokdoc(13)))                                       %(13.51b)
pause;

disp('Example 13.52 ----------------------------------------------------');
stpwrd = "(?:"+join(stpwds,'|')+")";                               %(13.52)
relwrd = "(?:are the|are|is the|is a|is|consists of)";
tokwrds = "([\w-]+ )*?";
pattern4 = entwrd+"\W+?"+relwrd+"\W+?"+tokwrds+stpwrd+"\W+?";
pause;

disp('Example 13.53 ----------------------------------------------------');
counter = 0;                                                      %(13.53a)
for k = 1:length(tokdoc)
    sentence = lower(joinWords(tokdoc(k)));
    [init, matches] = regexp(sentence,pattern4,'start','tokens');
    if isempty(matches), continue; end
    entity = matches{1}(1);
    tokens = split(strip(matches{1}(2)));
    for m = 1:length(tokens)
    % extracts the definition only if at least one of the tokens is a noun
        if sum(tokens(m)==lower(nouns))>0
            counter = counter+1;
            definitions(counter).sentn = k;
            definitions(counter).entity = entity;
            definition = string(sentence{1}(init(1):end));
            definitions(counter).definition = definition;
            break
        end
    end
end
whos definitions                                                  %(13.53b)
pause;

% PRINTS TABLE 13.3
struct2table(definitions)
pause

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
